"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
var __asyncDelegator = (this && this.__asyncDelegator) || function (o) {
    var i, p;
    return i = {}, verb("next"), verb("throw", function (e) { throw e; }), verb("return"), i[Symbol.iterator] = function () { return this; }, i;
    function verb(n, f) { i[n] = o[n] ? function (v) { return (p = !p) ? { value: __await(o[n](v)), done: n === "return" } : f ? f(v) : v; } : f; }
};
var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var g = generator.apply(thisArg, _arguments || []), i, q = [];
    return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
    function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
    function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }
    function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }
    function fulfill(value) { resume("next", value); }
    function reject(value) { resume("throw", value); }
    function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.walkAsync = exports.walk = exports.walkAsyncStep = exports.walkStep = void 0;
const defaults_1 = require("./defaults");
const helpers_1 = require("./helpers");
const utils_1 = require("./utils");
const node_1 = require("./node");
const callback_1 = require("./callback");
function shouldSkipVisitation(node, ctx) {
    if (node.nodeType === 'value')
        return false;
    const seen = ctx.seenObjects.indexOf(node.val) !== -1;
    if (seen && ctx.config.graphMode === 'graph')
        return true;
    if (seen && ctx.config.graphMode === 'finiteTree')
        throw "The object violates the defined structure. Override 'graphMode' in the config to allow parsing different object structures.";
    ctx.seenObjects.push(node.val);
    return false;
}
function* processNode(node, ctx, processChildren) {
    if (shouldSkipVisitation(node, ctx))
        return;
    callback_1.execCallbacks(callback_1.matchCallbacks(node, 'preWalk', ctx), node);
    yield node;
    if (processChildren)
        for (let child of node.getChildren())
            yield* processNode(child, ctx, true);
    callback_1.execCallbacks(callback_1.matchCallbacks(node, 'postWalk', ctx), node);
}
function processAsync(node, ctx, processChildren) {
    return __asyncGenerator(this, arguments, function* processAsync_1() {
        if (shouldSkipVisitation(node, ctx))
            return yield __await(void 0);
        const executor = callback_1.getAsyncExecutor(ctx.config);
        yield __await(executor(callback_1.matchCallbacks(node, 'preWalk', ctx), node));
        yield yield __await(node);
        if (processChildren)
            for (let child of node.getChildren())
                yield __await(yield* __asyncDelegator(__asyncValues(yield __await(processAsync(child, ctx, true)))));
        yield __await(executor(callback_1.matchCallbacks(node, 'postWalk', ctx), node));
    });
}
function buildContext(config) {
    const ctx = defaults_1.buildDefaultContext(config);
    ctx.config.callbacks.forEach((cb) => {
        const callback = Object.assign(Object.assign({}, cb), { executionOrder: typeof cb.executionOrder == 'undefined' ? 0 : cb.executionOrder, positionFilter: cb.positionFilter || "preWalk" });
        if (callback.positionFilter === "both") {
            ctx.callbacksByPosition["preWalk"].push(callback);
            ctx.callbacksByPosition["postWalk"].push(callback);
        }
        else if (typeof callback.positionFilter !== 'undefined') {
            ctx.callbacksByPosition[callback.positionFilter].push(callback);
        }
    });
    for (const key in ctx.callbacksByPosition)
        ctx.callbacksByPosition[key] = ctx.callbacksByPosition[key].sort(helpers_1.executionOrderSort);
    return ctx;
}
function* walkStep(obj, config) {
    const context = buildContext(config);
    const rootNode = node_1.WalkNode.fromRoot(obj);
    try {
        if (context.config.traversalMode === 'depth')
            yield* processNode(rootNode, context, true);
        else {
            let queue = [rootNode];
            do {
                const next = queue.shift();
                yield* processNode(next, context, false);
                queue.push(...next.children);
            } while (queue.length > 0);
        }
    }
    catch (err) {
        if (!(err instanceof utils_1.Break))
            throw err;
    }
}
exports.walkStep = walkStep;
function walkAsyncStep(obj, config) {
    return __asyncGenerator(this, arguments, function* walkAsyncStep_1() {
        const context = buildContext(config);
        const rootNode = node_1.WalkNode.fromRoot(obj);
        try {
            if (context.config.traversalMode === 'depth') {
                yield __await(yield* __asyncDelegator(__asyncValues(yield __await(processAsync(rootNode, context, true)))));
            }
            else {
                let queue = [rootNode];
                do {
                    const next = queue.shift();
                    yield __await(yield* __asyncDelegator(__asyncValues(yield __await(processAsync(next, context, false)))));
                    queue.push(...next.children);
                } while (queue.length > 0);
            }
        }
        catch (err) {
            if (!(err instanceof utils_1.Break))
                throw err;
        }
    });
}
exports.walkAsyncStep = walkAsyncStep;
const walk = (obj, config) => {
    utils_1.forceEvalGenerator(walkStep(obj, config));
};
exports.walk = walk;
const walkAsync = (obj, config) => __awaiter(void 0, void 0, void 0, function* () {
    yield utils_1.forceEvalAsyncGenerator(walkAsyncStep(obj, config));
});
exports.walkAsync = walkAsync;
